# -*- coding: utf-8 -*-

import numpy as np

from emcee.moves.red_blue import RedBlueMove

__all__ = ["StretchMove"]


class StretchMove(RedBlueMove):
    """
    A `Goodman & Weare (2010)
    <https://msp.org/camcos/2010/5-1/p04.xhtml>`_ "stretch move" with
    parallelization as described in `Foreman-Mackey et al. (2013)
    <https://arxiv.org/abs/1202.3665>`_.
    :param a: (optional)
        The stretch scale parameter. (default: ``2.0``)
    """

    def __init__(self, a=2.0, n=0, **kwargs):
        self.a = a
        self.n = n
        super(StretchMove, self).__init__(**kwargs)

    def setup(self, coords):
        self.a0 = self.a * 1.3**self.n

    def get_proposal(self, s, c, random):
        c = np.concatenate(c, axis=0)
        Ns, Nc = len(s), len(c)
        ndim = s.shape[1]
        zz = ((self.a0 - 1.0) * random.rand(Ns) + 1) ** 2.0 / self.a0
        factors = (ndim - 1.0) * np.log(zz)
        rint = random.randint(Nc, size=(Ns,))
        return c[rint] - (c[rint] - s) * zz[:, None], factors



import numpy

def sse(exp_data, sim_data):
    return numpy.sum( (numpy.array(exp_data) - numpy.array(sim_data))**2 )

def mse(exp_data, sim_data):
    return sse(exp_data, sim_data)/len(exp_data)

def rmse(exp_data, sim_data):
    return numpy.sqrt(mse(exp_data, sim_data))

def sse_norm(exp_data, sim_data):
    return numpy.sum( (numpy.array(exp_data) - numpy.array(sim_data))**2 )/(numpy.max(exp_data)**2)

def mse_norm(exp_data, sim_data):
    return sse_norm(exp_data, sim_data)/len(exp_data)

def norm_rmse(exp_data, sim_data):
    return numpy.sqrt(mse_norm(exp_data, sim_data))

def rmse_combine(exp_datas, sim_datas):
    len_entries = numpy.array([len(exp_data) for exp_data in exp_datas])
    total_entries = numpy.sum(len_entries)
    sse_s = numpy.array([sse_norm(exp_data, sim_data) for exp_data,sim_data in zip(exp_datas, sim_datas)])
    mse_s = sse_s/len_entries
    mean_total = numpy.sum(mse_s)    
    rmse = numpy.sqrt(mean_total)
    return rmse



import CADETMatch.util as util
import CADETMatch.checkpoint_algorithms as checkpoint_algorithms
import random

import CADETMatch.pareto as pareto
import array

import scipy.special
from deap import tools
import numpy

import multiprocessing

name = "NSGA3"

def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()
    parameters = len(cache.MIN_VALUE)

    populationSize=parameters * cache.settings.get('population', 100)
    CXPB = cache.settings.get('crossoverRate', 1.0)
    MUTPB = cache.settings.get('mutationRate', 1.0)

    totalGenerations = parameters * cache.settings.get('generations', 1000)

    pop = cache.toolbox.population(n=populationSize)

    if "seeds" in cache.settings:
        seed_pop = [cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings['transform'], sublist)]) for sublist in cache.settings['seeds']]
        pop.extend(seed_pop)
    return checkpoint_algorithms.eaMuPlusLambda(pop, cache.toolbox,
                              mu=populationSize, 
                              lambda_=populationSize, 
                              cxpb=CXPB, 
                              mutpb=MUTPB,
                              ngen=totalGenerations,
                              settings=cache.settings,
                              tools=tools,
                              cache=cache)

def setupDEAP(cache, fitness, fitness_final, grad_fitness, grad_search, grad_search_fine, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None,
                   csv_line=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None,
                   csv_line=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register("individual", util.generateIndividual, creator.Individual,
        len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache)
        
    if cache.sobolGeneration:
        cache.toolbox.register("population", util.sobolGenerator, creator.Individual, cache)
    else:
        cache.toolbox.register("population", tools.initRepeat, list, cache.toolbox.individual)
    cache.toolbox.register("randomPopulation", tools.initRepeat, list, cache.toolbox.individual)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register("mate", tools.cxSimulatedBinaryBounded, eta=30.0, low=cache.MIN_VALUE, up=cache.MAX_VALUE)

    cache.toolbox.register("mutate", tools.mutPolynomialBounded, eta=20.0, low=cache.MIN_VALUE, up=cache.MAX_VALUE, indpb=1.0/len(cache.MIN_VALUE))
    cache.toolbox.register("force_mutate", tools.mutPolynomialBounded, eta=20.0, low=cache.MIN_VALUE, up=cache.MAX_VALUE, indpb=1.0/len(cache.MIN_VALUE))

    if cache.numGoals == 1:
        #NSGA3 uses reference points and is not suitable for a single objective, switch to NSGA2
        cache.toolbox.register("select", tools.selNSGA2)
    else:
        ref_points = generate_reference_points(cache.numGoals)
        cache.toolbox.register("select", tools.selNSGA3WithMemory(ref_points))

    cache.toolbox.register("evaluate", fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_final", fitness_final, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad", grad_fitness, json_path=cache.json_path)
    cache.toolbox.register("evaluate_grad_fine", grad_search_fine, json_path=cache.json_path)
    cache.toolbox.register('grad_search', grad_search)

    cache.toolbox.register('map', map_function)


def num_ref_points(n,k):
    return scipy.special.comb(n+k-1,k, exact=False)

def find_max_p(ndim, max_size, max_p):
    for p in range(max_p, 0, -1):
        if num_ref_points(ndim, p) <= max_size:
            break
    return p

def find_ref_point_setup(ndim, max_size):
    points = [(k,num_ref_points(ndim,k)) for k in range(128, 0, -1)]
    size = 0
    P = []
    for p, p_size in points:
        if p_size <= (max_size - size):
            P.append(p)
            size += p_size
    S = [1/2.0**n for n in range(len(P))]
    return P, S

def generate_reference_points(ndim, max_size=100):
    P, SCALES = find_ref_point_setup(ndim, max_size)
    ref_points = [tools.uniform_reference_points(ndim, p, s) for p, s in zip(P, SCALES)]
    ref_points = numpy.concatenate(ref_points, axis=0)
    _, uniques = numpy.unique(ref_points, axis=0, return_index=True)
    ref_points = ref_points[uniques]
    multiprocessing.get_logger().info("Reference points chosen P = %s  with shape %s", P, ref_points.shape)
    return ref_points


from pathlib import Path
import pickle
import random
import numpy
import CADETMatch.util as util
from deap import algorithms
import time
import csv
import CADETMatch.pareto as pareto
import multiprocessing

stallRate = 1.25
progressRate = 0.75

def eaMuCommaLambda(population, toolbox, mu, lambda_, cxpb, mutpb, ngen, settings,
                    stats=None, verbose=__debug__, tools=None, cache=None):
    """from DEAP function but with checkpoiting"""

    assert lambda_ >= mu, "lambda must be greater or equal to mu."

    checkpointFile = Path(settings['resultsDirMisc'], settings.get('checkpointFile', 'check'))

    sim_start = generation_start = time.time()

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[]}
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        if checkpointFile.exists():
            with checkpointFile.open('rb') as cp_file:
                cp = pickle.load(cp_file)
            population = cp["population"]
            start_gen = cp["generation"]    
    
            halloffame = cp["halloffame"]
            meta_hof = cp['meta_halloffame']
            grad_hof = cp['grad_halloffame']
            progress_hof = cp['progress_halloffame']
            random.setstate(cp["rndstate"])
            cache.generationsOfProgress = cp['generationsOfProgress']
            cache.lastProgressGeneration = cp['lastProgressGeneration']

            if cp['gradCheck'] > cache.settings.get('gradCheck', 1.0):
                gradCheck = cp['gradCheck']
            else:
                gradCheck = cache.settings.get('gradCheck', 1.0)

        else:
            # Start a new evolution
            start_gen = 0    

            gradCheck = settings.get('gradCheck', 1.0)

            if cache.metaResultsOnly:
                halloffame = pareto.DummyFront()
            else:
                halloffame = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            meta_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
            grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            progress_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))


            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in population if not ind.fitness.valid]
            if invalid_ind:
                stalled, stallWarn, progressWarn = util.eval_population(toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                        progress_hof, -1, result_data)

                util.writeProgress(cache, -1, population, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
                util.graph_process(cache, "First")
                util.graph_corner_process(cache, last=False)

            cp = dict(population=population, generation=start_gen, halloffame=halloffame,
                rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame = progress_hof)

            with checkpointFile.open('wb')as cp_file:
                pickle.dump(cp, cp_file)

        # Begin the generational process
        for gen in range(start_gen, ngen+1):
            generation_start = time.time()
            # Vary the population
            offspring = algorithms.varOr(population, toolbox, lambda_, cxpb, mutpb)

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            stalled, stallWarn, progressWarn = util.eval_population(toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                    progress_hof, gen, result_data)

            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, gen)
            offspring.extend(newChildren)

            # Select the next generation population
            population[:] = toolbox.select(offspring, mu)

            util.writeProgress(cache, gen, offspring, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
            util.graph_process(cache, gen)
            util.graph_corner_process(cache, last=False)

            if stallWarn:
                maxPopulation = cache.settings['maxPopulation'] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * stallRate)
                lambda_ = min(newLambda_, maxPopulation)

            if progressWarn:
                minPopulation = cache.settings['minPopulation'] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * progressRate)
                lambda_ = max(newLambda_, minPopulation)
                                   
            cp = dict(population=population, generation=gen, halloffame=halloffame,
                rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame = progress_hof)

            with checkpointFile.open('wb') as cp_file:
                pickle.dump(cp, cp_file)

            if avg >= settings.get('stopAverage', 1.0) or bestMin >= settings.get('stopBest', 1.0) or stalled:
                util.finish(cache)
                util.graph_corner_process(cache, last=True)
                return halloffame
        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return halloffame


def eaMuPlusLambda(population, toolbox, mu, lambda_, cxpb, mutpb, ngen, settings,
                   stats=None, verbose=__debug__, tools=None, cache=None):
    """from DEAP function but with checkpoiting"""
    assert lambda_ >= mu, "lambda must be greater or equal to mu."

    checkpointFile = Path(settings['resultsDirMisc'], settings.get('checkpointFile', 'check'))

    sim_start = generation_start = time.time()

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[]}
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        if checkpointFile.exists():
            with checkpointFile.open('rb') as cp_file:
                cp = pickle.load(cp_file)
            population = cp["population"]
            start_gen = cp["generation"]    
    
            halloffame = cp["halloffame"]
            meta_hof = cp['meta_halloffame']
            grad_hof = cp['grad_halloffame']
            progress_hof = cp['progress_halloffame']
            random.setstate(cp["rndstate"])
            cache.generationsOfProgress = cp['generationsOfProgress']
            cache.lastProgressGeneration = cp['lastProgressGeneration']
            
            if cp['gradCheck'] > cache.settings.get('gradCheck', 1.0):
                gradCheck = cp['gradCheck']
            else:
                gradCheck = cache.settings.get('gradCheck', 1.0)
        else:
            # Start a new evolution
            start_gen = 0    

            gradCheck = settings.get('gradCheck', 1.0)

            if cache.metaResultsOnly:
                halloffame = pareto.DummyFront()
            else:
                halloffame = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            meta_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
            grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            progress_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in population if not ind.fitness.valid]
            if invalid_ind:
                stalled, stallWarn, progressWarn = util.eval_population(toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                        progress_hof, -1, result_data)

                util.writeProgress(cache, -1, population, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
                util.graph_process(cache, "First")
                util.graph_corner_process(cache, last=False)

            cp = dict(population=population, generation=start_gen, halloffame=halloffame,
                rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame = progress_hof)

            with checkpointFile.open('wb')as cp_file:
                pickle.dump(cp, cp_file)

        gen = start_gen  #this covers the case where the start_gen is higher than our stop gen so the loop never runs
        # Begin the generational process
        for gen in range(start_gen, ngen+1):
            generation_start = time.time()
            # Vary the population
            offspring = algorithms.varAnd(population, toolbox, cxpb, mutpb)

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            stalled, stallWarn, progressWarn = util.eval_population(toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                    progress_hof, gen, result_data)
                        
            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, gen)
            offspring.extend(newChildren)

            # Select the next generation population
            population[:] = toolbox.select(offspring + population, mu)

            util.writeProgress(cache, gen, offspring, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
            util.graph_process(cache, gen)
            util.graph_corner_process(cache, last=False)

            cp = dict(population=population, generation=gen, halloffame=halloffame,
                rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame = progress_hof)

            if stallWarn:
                maxPopulation = cache.settings['maxPopulation'] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * stallRate)
                lambda_ = min(newLambda_, maxPopulation)

            if progressWarn:
                minPopulation = cache.settings['minPopulation'] * len(cache.MIN_VALUE)
                newLambda_ = int(lambda_ * progressRate)
                lambda_ = max(newLambda_, minPopulation)


            with checkpointFile.open('wb') as cp_file:
                pickle.dump(cp, cp_file)

            best = meta_hof.getBestScores()

            if best[2] >= settings.get('stopAverage', 1.0) or best[1] >= settings.get('stopBest', 1.0) or stalled:
                break

        if cache.finalGradRefinement:
            gen = gen + 1
            best_individuals = [cache.toolbox.individual_guess(i) for i in meta_hof]
            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, best_individuals, cache, writer, csvfile, 
                                                               grad_hof, meta_hof, gen, check_all=True, result_data=result_data)
            if newChildren:
                util.writeProgress(cache, gen, newChildren, halloffame, meta_hof, grad_hof,
                                   sim_start, generation_start, result_data)

        population = [cache.toolbox.individual_guess(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(cache.toolbox, cache, population, writer, csvfile, halloffame, meta_hof, 
                                                                      progress_hof, gen+1, result_data)
        util.writeProgress(cache, gen+1, population, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)

        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return halloffame

def nsga2(populationSize, ngen, cache, tools):
    """NSGA2 with checkpointing"""
    #fix max population to be a multiple of 4
    cache.settings['maxPopulation'] = cache.settings['maxPopulation'] + (-cache.settings['maxPopulation'] % 4)

    cxpb = cache.settings['crossoverRate']
    checkpointFile = Path(cache.settings['resultsDirMisc'], cache.settings.get('checkpointFile', 'check'))

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[]}
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        if checkpointFile.exists():
            with checkpointFile.open('rb') as cp_file:
                cp = pickle.load(cp_file)
            population = cp["population"]
            start_gen = cp["generation"]    
    
            halloffame = cp["halloffame"]
            meta_hof = cp['meta_halloffame']
            grad_hof = cp['grad_halloffame']
            progress_hof = cp['progress_halloffame']
            random.setstate(cp["rndstate"])
            cache.generationsOfProgress = cp['generationsOfProgress']
            cache.lastProgressGeneration = cp['lastProgressGeneration']

            if cp['gradCheck'] > cache.settings.get('gradCheck', 1.0):
                gradCheck = cp['gradCheck']
            else:
                gradCheck = cache.settings.get('gradCheck', 1.0)

        else:
            # Start a new evolution

            population = cache.toolbox.population(n=populationSize)

            if "seeds" in cache.settings:
                seed_pop = [cache.toolbox.individual_guess([f(v) for f, v in zip(cache.settings['transform'], sublist)]) for sublist in cache.settings['seeds']]
                population.extend(seed_pop)

            start_gen = 0    

            if cache.metaResultsOnly:
                halloffame = pareto.DummyFront()
            else:
                halloffame = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            meta_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
            grad_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit(cache))
            progress_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))
            gradCheck = cache.settings.get('gradCheck', 1.0)


        sim_start = generation_start = time.time()
   
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        if invalid_ind:
            stalled, stallWarn, progressWarn = util.eval_population(cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                    progress_hof, -1, result_data)
            
            util.writeProgress(cache, -1, population, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
            util.graph_process(cache, "First")
            util.graph_corner_process(cache, last=False)
        
        # This is just to assign the crowding distance to the individuals
        # no actual selection is done
        population = cache.toolbox.select(population, len(population)) 
    

        cp = dict(population=population, generation=start_gen, halloffame=halloffame,
            rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
            generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
            progress_halloffame = progress_hof)

        with checkpointFile.open('wb')as cp_file:
            pickle.dump(cp, cp_file)

        # Begin the generational process
        for gen in range(start_gen, ngen+1):
            generation_start = time.time()
            # Vary the population
            offspring = tools.selTournamentDCD(population, len(population))
            offspring = [cache.toolbox.clone(ind) for ind in offspring]
                
            for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
                if random.random() <= cxpb:
                    cache.toolbox.mate(ind1, ind2)
            
                cache.toolbox.mutate(ind1)
                cache.toolbox.mutate(ind2)
                del ind1.fitness.values, ind2.fitness.values
        
            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            stalled, stallWarn, progressWarn = util.eval_population(cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                                                    progress_hof, gen, result_data)

            gradCheck, newChildren = cache.toolbox.grad_search(gradCheck, offspring, cache, writer, csvfile, grad_hof, meta_hof, gen)
            offspring.extend(newChildren)

            util.writeProgress(cache, gen, offspring, halloffame, meta_hof, grad_hof, sim_start, generation_start, result_data)
            util.graph_process(cache, gen)
            util.graph_corner_process(cache, last=False)

            # Select the next generation population
            population = cache.toolbox.select(population + offspring, populationSize)

            if stallWarn:
                maxPopulation = cache.settings['maxPopulation'] * len(cache.MIN_VALUE)
                newPopulationSize = int(populationSize * stallRate)
                newPopulationSize += (-newPopulationSize % 4)
                newPopulationSize = min(newPopulationSize, maxPopulation)
                newPopulationSize += (-newPopulationSize % 4)

                diffSize = newPopulationSize - populationSize
                newPopulation = cache.toolbox.randomPopulation(n=diffSize)

                invalid_ind = [ind for ind in newPopulation if not ind.fitness.valid]
                util.eval_population(cache.toolbox, cache, invalid_ind, writer, csvfile, halloffame, meta_hof, 
                                     progress_hof, gen, result_data)

                # This is just to assign the crowding distance to the individuals
                # no actual selection is done
                newPopulation = cache.toolbox.select(newPopulation, len(newPopulation)) 
                
                population.extend(newPopulation)
                populationSize = newPopulationSize

            if progressWarn:
                minPopulation = cache.settings['minPopulation'] * len(cache.MIN_VALUE)
                newPopulationSize = int(populationSize * progressRate)
                newPopulationSize += (-newPopulationSize % 4)
                newPopulationSize = max(newPopulationSize, minPopulation)
                newPopulationSize += (-newPopulationSize % 4)

                diffSize = populationSize - newPopulationSize
                
                population = cache.toolbox.select(population, newPopulationSize) 
                populationSize = newPopulationSize

            cp = dict(population=population, generation=gen, halloffame=halloffame,
                rndstate=random.getstate(), gradCheck=gradCheck, meta_halloffame=meta_hof, grad_halloffame=grad_hof,
                generationsOfProgress=cache.generationsOfProgress, lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame = progress_hof)

            hof = Path(cache.settings['resultsDirMisc'], 'hof')
            with hof.open('wb') as data:
                numpy.savetxt(data, numpy.array(halloffame))
            with checkpointFile.open('wb') as cp_file:
                pickle.dump(cp, cp_file)

            best = meta_hof.getBestScores()

            if best[2] >= cache.settings.get('stopAverage', 1.0) or best[1] >= cache.settings.get('stopBest', 1.0) or stalled:
                util.finish(cache)
                util.graph_corner_process(cache, last=True)
                return halloffame
        util.finish(cache)
        util.graph_corner_process(cache, last=True)
        return halloffame



import sys

class LoggerWriter:
    def __init__(self, level):
        self.level = level

    def write(self, message):
        if message != '\n':
            self.level(message)

    def flush(self):
        pass
        #self.level(sys.stderr)


import sys
from CADETMatch.cache import cache
from pathlib import Path

import matplotlib
matplotlib.use("Agg")

import matplotlib.pyplot as plt
import matplotlib.animation as animation

def main():
    cache.setup(sys.argv[1])
    cache.progress_path = Path(cache.settings['resultsDirBase']) / "progress.csv"

    output_spear = cache.settings['resultsDirSpace'] / "spearman"

    pngs = list(output_spear.glob('*.png'))

    pngs_sorted = sorted( pngs, key = lambda file: file.stat().st_ctime)

    fig = plt.figure(figsize=(10,10), dpi=100)

    ims = []
    for png in pngs_sorted:
        with png.open('rb') as image_file:
            data = plt.imread(image_file)
        ymax, xmax, _ = data.shape
        im = plt.imshow(data, animated=True, aspect="equal", extent=[0,xmax,ymax,0])
        plt.axis('off')
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=200, blit=False,
                                repeat=False)

    FFMpegWriter = animation.writers['ffmpeg']
    writer = FFMpegWriter(fps=5, bitrate=50000)
    fig.tight_layout()

    ani.save(str(output_spear / 'spearman.mp4'), writer=writer)

if __name__ == "__main__":
    main()




import sys

import matplotlib
matplotlib.use('Agg')

size = 20

matplotlib.rc('font', size=size)          # controls default text sizes
matplotlib.rc('axes', titlesize=size)     # fontsize of the axes title
matplotlib.rc('axes', labelsize=size)    # fontsize of the x and y labels
matplotlib.rc('xtick', labelsize=size)    # fontsize of the tick labels
matplotlib.rc('ytick', labelsize=size)    # fontsize of the tick labels
matplotlib.rc('legend', fontsize=size)    # legend fontsize
matplotlib.rc('figure', titlesize=size)  # fontsize of the figure title
matplotlib.rc('figure', autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D  

from CADETMatch.cache import cache

from pathlib import Path
import pandas
import numpy
import scipy.interpolate
import itertools

from cadet import H5
from addict import Dict

#parallelization
import multiprocessing

import os
import warnings
import corner
import seaborn as sns
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter

saltIsotherms = {b'STERIC_MASS_ACTION', b'SELF_ASSOCIATION', b'MULTISTATE_STERIC_MASS_ACTION', 
                 b'SIMPLE_MULTISTATE_STERIC_MASS_ACTION', b'BI_STERIC_MASS_ACTION'}


from matplotlib.colors import ListedColormap
import matplotlib.cm
cmap = matplotlib.cm.winter

# Get the colormap colors
my_cmap = cmap(numpy.arange(cmap.N))

# Set alpha
my_cmap[:,-1] = 0.05

# Create new colormap
my_cmap = ListedColormap(my_cmap)

cm_plot = matplotlib.cm.gist_rainbow

def new_range(flat_chain):
    lb_data, mid_data, ub_data = numpy.percentile(flat_chain, [5, 50, 95], 0)
    
    distance = numpy.max(numpy.abs(numpy.array([lb_data - mid_data, ub_data - mid_data])), axis=0)
    
    lb_range = mid_data - 2 * distance
    ub_range = mid_data + 2 * distance
    
    return numpy.array([lb_range, ub_range])

def get_color(idx, max_colors, cmap):
    return cmap(1.*float(idx)/max_colors)

def main():    
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings['resultsDirLog'], "graph_corner.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    graphCorner(cache)

def plotChain(flat_chain, flat_chain_transform, headers, out_dir, prefix):
    multiprocessing.get_logger().info("plotting chain")
    if len(flat_chain) > 1e5:
        indexes = numpy.random.choice(flat_chain.shape[0], int(1e5), replace=False)
        chain = flat_chain[indexes]
        chain_transform = flat_chain_transform[indexes]
    else:
        chain = flat_chain
        chain_transform = flat_chain_transform
    
    #stat corner plot
    fig_size = 6 * len(headers)
    fig = corner.corner(chain, quantiles=(0.16, 0.5, 0.84),
                   show_titles=True, labels=headers, 
                    bins=20, range=new_range(chain).T, 
                     use_math_text=True, title_fmt='.2g')    
    fig.set_size_inches((fig_size,fig_size))
    fig.savefig(str(out_dir / ("%s_corner.png" % prefix)))

    fig = corner.corner(chain_transform, quantiles=(0.16, 0.5, 0.84),
                   show_titles=True, labels=headers, 
                    bins=20, range=new_range(chain_transform).T, 
                     use_math_text=True, title_fmt='.2g') 
    fig.set_size_inches((fig_size,fig_size))
    fig.savefig(str(out_dir / ("%s_corner_transform.png" % prefix)))

def plotMCMCParam(out_dir, param, chain, header):
    multiprocessing.get_logger().info('plotting mcmc param %s', param)
    shape = chain.shape
    x = numpy.linspace(0, shape[1], shape[1])

    fig = figure.Figure(figsize=[10, 10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    graph.plot(x, chain[0,:,param], 'r', label='5')
    graph.plot(x, chain[1,:,param], 'k', label='mean')
    graph.plot(x, chain[2,:,param], 'r', label='95')
    graph.fill_between(x, chain[0,:,param], chain[2,:,param], facecolor='r', alpha=0.5)
    graph.legend()
    graph.set_title(header)
    fig.set_size_inches((12,12))
    fig.savefig(str(out_dir / ("%s.png" % header) ))

def plotMCMCVars(out_dir, headers, data):
    for param_idx, header in enumerate(headers):
        if 'train_chain_stat' in data.root:
            plotMCMCParam(out_dir, param_idx, data.root.train_chain_stat, "Train " + header + str(param_idx))
            plotMCMCParam(out_dir, param_idx, data.root.train_chain_stat_transform, "Train " + header + " Transform" + str(param_idx))

        if 'run_chain_stat' in data.root:
            plotMCMCParam(out_dir, param_idx, data.root.run_chain_stat, "Run " + header + str(param_idx))
            plotMCMCParam(out_dir, param_idx, data.root.run_chain_stat_transform, "Run " + header + " Transform" + str(param_idx))

def keep_header(char):
    allowed = set('-0123456789')
    return char in allowed

def clean_header(header):
    splits = header.split()
    temp = [splits[0]]
    for string in splits[1:]:
        temp.append(''.join([i for i in string if keep_header(i)]))
    return ' '.join(temp)

def graphCorner(cache):
    multiprocessing.get_logger().info("plotting corner plots")
    headers = list(cache.parameter_headers_actual)
    headers = [clean_header(header) for header in headers]
    
    resultDir = Path(cache.settings['resultsDir'])
    result_h5 = resultDir / "result.h5"

    miscDir = Path(cache.settings['resultsDirMCMC'])
    mcmc_h5 = miscDir / "mcmc.h5"

    if mcmc_h5.exists():
        data = H5()
        data.filename = mcmc_h5.as_posix()
        data.load()

        out_dir = cache.settings['resultsDirProgress']

        if 'tau_percent' in data.root:
            multiprocessing.get_logger().info('plotting tau plots')
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            tau_percent = numpy.squeeze(data.root.tau_percent)
            max_values = numpy.ones(tau_percent.shape)
            tau_percent = numpy.min([tau_percent, max_values], 0)

            graph.bar(list(range(len(headers))), tau_percent, tick_label = headers)
            fig.set_size_inches((12,12))
            fig.savefig(str(out_dir / "tau_percent.png" ))

        if 'flat_chain' in data.root:
            multiprocessing.get_logger().info('plot flat chain')
            plotChain(data.root.flat_chain, data.root.flat_chain_transform, headers, out_dir, 'full')

        if 'burn_in_acceptance' in data.root:
            multiprocessing.get_logger().info('plot burn in acceptance')
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            graph.plot(util.get_confidence(data.root.burn_in_acceptance))
            graph.set_title("Burn In Acceptance")
            graph.set_xlabel('Step')
            graph.set_ylabel('Acceptance')
            fig.set_size_inches((12,12))
            fig.savefig(str(out_dir / "burn_in_acceptance.png"))

        if 'integrated_autocorrelation_time' in data.root:
            iat = data.root.integrated_autocorrelation_time

            multiprocessing.get_logger().info('plot integrated autocorrelation time')
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            for i in range(1, iat.shape[1]):
                graph.plot(iat[:,0], iat[:,i], label=i)
            graph.set_title("Integrated Autocorrelation Time")
            graph.set_xlabel('Step')
            graph.set_ylabel('IAT')
            graph.legend()
            fig.set_size_inches((12,12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time.png"))

            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            tau = cache.MCMCTauMult
            for i in range(1, iat.shape[1]):
                progress = (iat[:,0]/tau)/iat[:,i]
                progress[progress > 1] = 1
                graph.plot(iat[:,0], progress, label=i)
            graph.set_title("Integrated Autocorrelation Time Progress")
            graph.set_xlabel('Step')
            graph.set_ylabel('Progress')
            graph.legend()
            fig.set_size_inches((12,12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time_progress.png"))

        if 'mcmc_acceptance' in data.root:
            multiprocessing.get_logger().info('plot mcmc acceptance')
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            graph.plot(util.get_confidence(data.root.mcmc_acceptance))
            graph.set_title("MCMC Acceptance")
            graph.set_xlabel('Step')
            graph.set_ylabel('Acceptance')
            fig.set_size_inches((12,12))
            fig.savefig(str(out_dir / "mcmc_acceptance.png"))

        plotChain(data.root.train_flat_chain, data.root.train_flat_chain_transform, headers, out_dir, 'train')
        plotMCMCVars(out_dir, headers, data)

    else:
        data = H5()
        data.filename = result_h5.as_posix()
        data.load()

        if len(data.root.input) > 1e5:
            indexes = numpy.random.choice(data.root.input.shape[0], int(1e5), replace=False)
            data_input = data.root.input[indexes]
            data_input_transform = data.root.input_transform[indexes]
            weight_min = data.root.output_meta[indexes,1]
            weight_prod = data.root.output_meta[indexes,2]
            all_scores = data.root.output[indexes]
        else:
            data_input = data.root.input
            data_input_transform = data.root.input_transform
            weight_min = data.root.output_meta[:,1]
            weight_prod = data.root.output_meta[:,2]
            all_scores = data.root.output

        max_scores = numpy.max(all_scores, 1)
        acceptable = max_scores > 0.01

        if numpy.any(acceptable):
            multiprocessing.get_logger().info("graphing remove %s points", len(max_scores) - numpy.sum(acceptable))

            data_input = data_input[acceptable]
            data_input_transform = data_input_transform[acceptable]
            weight_min = weight_min[acceptable]
            weight_prod = weight_prod[acceptable]

        out_dir = cache.settings['resultsDirProgress']

        create_corner(out_dir, "corner.png", headers, data_input, weights=None)
        create_corner(out_dir, "corner_min.png", headers, data_input, weights=weight_min)
        create_corner(out_dir, "corner_prod.png", headers, data_input, weights=weight_prod)

        #transformed entries
        create_corner(out_dir, "corner_transform.png", headers, data_input_transform, weights=None)
        create_corner(out_dir, "corner_min_transform.png", headers, data_input_transform, weights=weight_min)
        create_corner(out_dir, "corner_prod_transform.png", headers, data_input_transform, weights=weight_prod)

def create_corner(dir, filename, headers, data, weights=None):
    if len(data) <= len(headers):
        multiprocessing.get_logger().info("There is not enough data to generate corner plots")
        return
    if  numpy.all(numpy.min(data,0) < numpy.max(data,0)):
        if weights is None or numpy.max(weights) > numpy.min(weights):
            bounds = new_range(data).T

            fig_size = 6 * len(headers)
            fig = corner.corner(data, quantiles=(0.16, 0.5, 0.84),
                   show_titles=True, labels=headers, 
                    bins=20, range=new_range(data).T, 
                     use_math_text=True, title_fmt='.2g')
            fig.set_size_inches((fig_size,fig_size))
            fig.savefig(str(dir / filename))        

            df = pandas.DataFrame(data, columns=headers)
            fig = sns.PairGrid(df, diag_sharey=False, height=fig_size)
            fig.map_lower(sns.kdeplot)
            fig.map_upper(sns.scatterplot)
            fig.map_diag(sns.kdeplot, lw=3)

            for row_num, row in enumerate(bounds):
                for col_num, col in enumerate(bounds):
                    if row_num == col_num:
                        fig.axes[row_num, col_num].set_xlim(*row)
                    else:
                        fig.axes[row_num, col_num].set_xlim(*col)
                        fig.axes[row_num, col_num].set_ylim(*row)

            fig.savefig(str(dir / ('sns_%s' % filename) ))

if __name__ == "__main__":
    main()




from deap import tools
from operator import eq
import math
import numpy

smallest = numpy.finfo(1.0).tiny

diff_step = 2.5e-2

class ParetoFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None, similar_fit = None):
        if similar is None:
            similar = eq
        if similar_fit is not None:
            self.similar_fit = similar_fit
        else:
            self.similar_fit = eq
        super().__init__(similar)

    def update(self, population):
        """Update the Pareto front hall of fame with the *population* by adding 
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.
        
        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        new_members = []
        significant = []
        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []
            for i, hofer in enumerate(self):    # hofer = hall of famer
                if not dominates_one and hofer.fitness.dominates(ind.fitness):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hofer.fitness):
                    dominates_one = True
                    to_remove.append(i)
                    significant.append(not self.similar_fit(ind.fitness.values, hofer.fitness.values))
                elif self.similar_fit(ind.fitness.values, hofer.fitness.values) and self.similar(ind, hofer):
                    has_twin = True
                    break
            
            for i in reversed(to_remove):       # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                self.insert(ind)
                new_members.append(ind)
                significant.append(True)
        return new_members, any(significant)

    def getBestScores(self):
        weights = numpy.array(self[0].fitness.weights)
        data_meta = numpy.array([i.fitness.values for i in self])
        return numpy.max(data_meta*weights, 0)*weights

class DummyFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None):
        "This is here for API compatibility, don't do anything"
        if similar is None:
            similar = eq
        super().__init__(similar)

    def update(self, population, live_mode=True):
        "do not put anything in this front, it is just needed to maintain compatibility"
        return [], False

def similar(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))
    
    #used to catch division by zero
    a[a == 0.0] = smallest

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_norm(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))

    #used to catch division by zero
    a[a == 0.0] = smallest

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_sse(a, b):
    "we only need a parameter to 4 digits of accuracy so have the pareto front only keep up to 5 digits for members of the front"
    a = numpy.absolute(numpy.array(a))
    b = numpy.absolute(numpy.array(b))

    #used to catch division by zero
    a[a == 0.0] = smallest

    a = numpy.log(a)
    b = numpy.log(b)

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_meta_norm(a, b):
    a = numpy.array(a)
    b = numpy.array(b)

    #used to catch division by zero
    a[a == 0.0] = smallest

    #SSE is in the last slot so we only want to use the first 3 meta scores
    a = a[:3]
    b = b[:3]

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_meta_sse(a, b):
    "SSE is negative and in the last slot and the only score needed"
    a = a[-2]
    b = b[-2]

    #used to catch division by zero
    if a == 0.0:
       a = smallest

    #SSE should be handled in log scale
    a = math.log(a)
    b = math.log(b)

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_meta_sse_split(a, b):
    "SSE is negative and in the last slot and the only score needed"
    a = numpy.abs(numpy.array(a))
    b = numpy.abs(numpy.array(b))

    #used to catch division by zero
    a[a == 0.0] = smallest

    #SSE should be handled in log scale
    a = numpy.log(a[:3])
    b = numpy.log(b[:3])

    diff = numpy.abs((a-b)/a)
    return numpy.all(diff < diff_step)

def similar_fit_meta(cache):
    if cache.allScoreNorm:
        return similar_fit_meta_norm
    elif cache.allScoreSSE:
        if cache.MultiObjectiveSSE:
            return similar_fit_meta_sse_split
        else:
            return similar_fit_meta_sse

def similar_fit(cache):
    if cache.allScoreNorm:
        return similar_fit_norm
    elif cache.allScoreSSE:
        return similar_fit_sse

def updateParetoFront(halloffame, offspring, cache):
    new_members, significant  = halloffame.update([offspring,])
    return bool(new_members), significant


import CADETMatch.util as util
import CADETMatch.pareto as pareto

from deap import creator
from deap import tools
from deap import base
import array

meta_hof = pareto.ParetoFront(similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache))

creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0, -1.0])
creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None,
                   csv_line=None)

toolbox = base.Toolbox()

def initIndividual(icls, content):
    return icls(content)

toolbox.register("individualMeta", initIndividual, creator.IndividualMeta)

ind_meta = toolbox.individualMeta([5.71E-07,	0.325681352])
ind_meta.fitness.values = [-0.297037238, -0.297037238, -0.297037238, 0.297037238]


new = meta_hof.update([ind_meta])

ind_meta = toolbox.individualMeta([5.73E-07,	0.325681352])
ind_meta.fitness.values = [-0.296955155, -0.296955155, -0.296955155, -0.296955155]

new = meta_hof.update([ind_meta])

a = None


import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import numpy
import numpy.linalg
from addict import Dict
import CADETMatch.smoothing as smoothing
import multiprocessing

name = "DextranShape"

def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 2
    return settings

def run(sim_data, feature):
    "special score designed for dextran. This looks at only the front side of the peak up to the maximum slope and pins a value at the elbow in addition to the top"
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)

    exp_time_zero = feature['exp_time_zero']
    exp_data_zero = feature['exp_data_zero']
    
    sim_spline, sim_data_zero_sse = cut_front(sim_time_values, sim_data_values, 
                                             feature['min_value_front'], feature['max_value_front'],
                                             feature['critical_frequency'], feature['smoothing_factor'])

    pearson, diff_time = score.pearson_spline_fun(exp_time_zero, exp_data_zero, sim_spline)

    exp_data_zero_sse = feature['exp_data_zero_sse']

    temp = [pearson,
            feature['offsetTimeFunction'](numpy.abs(diff_time)),
            ]

    data = (temp, util.sse(sim_data_zero_sse, exp_data_zero_sse), len(sim_data_zero_sse), 
            sim_time_values, sim_data_zero_sse, exp_data_zero_sse, [1.0 - i for i in temp])

    return data

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    #change the stop point to be where the max positive slope is along the searched interval
    name = '%s_%s' % (sim.root.experiment_name,   feature['name'])
    exp_time_zero, exp_data_zero, exp_data_zero_sse, min_time, min_value, max_time, max_value, s, crit_fs, crit_fs_der = cut_front_find(selectedTimes, selectedValues, name, cache)

    multiprocessing.get_logger().info("Dextran %s  start: %s   stop: %s  max value: %s", name, 
                                      min_time, max_time, max_value)

    temp['min_time'] = feature['start']
    temp['max_time'] = feature['stop']
    
    temp['min_time_front'] = min_time
    temp['min_value_front'] = min_value
    temp['max_time_front'] = max_time
    temp['max_value_front'] = max_value

    temp['exp_time_zero'] = exp_time_zero
    temp['exp_data_zero'] = exp_data_zero
    temp['exp_data_zero_sse'] = exp_data_zero_sse

    decay = feature.get('decay', 1)

    if decay:
        temp['offsetTimeFunction'] = score.time_function_cv(CV_time, selectedTimes, max_time)
    else:
        temp['offsetTimeFunction'] = score.time_function_decay_cv(CV_time, selectedTimes, max_time)

    temp['peak_max'] = max_value
    temp['smoothing_factor'] = s
    temp['critical_frequency'] = crit_fs
    temp['critical_frequency_der'] = crit_fs_der
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_Shape" % name, 
            "%s_Time" % name,
            ]
    return temp

def cut_front_find(times, values, name, cache):
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(times, values, name, cache)
    values_der = smoothing.smooth_data_derivative(times, values, crit_fs, s, crit_fs_der)

    smooth_value = smoothing.smooth_data(times, values, crit_fs, s)
    
    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times, values_der, ext=1)
    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, smooth_value, ext=1)

    min_time, min_value, max_time, max_value = score.find_cuts(times, smooth_value, spline, spline_der)
    
    #resample to 100 points/second
    needed_points = int( (times[-1] - times[0]) * 100)
    
    new_times = numpy.linspace(times[0], times[-1], needed_points)
    new_values = spline(new_times)

    exp_data_zero_sse, exp_min_time, exp_min_value, exp_max_time, exp_max_value = score.cut_zero(times, smooth_value, min_value, max_value)

    if exp_min_time is not None:
        min_time = exp_min_time
        min_value = exp_min_value
        max_time = exp_max_time
        max_value = exp_max_value

    exp_data_zero, _, _, _, _ = score.cut_zero(new_times, new_values, min_value, max_value)    
    
    return (new_times, exp_data_zero, exp_data_zero_sse, 
        min_time, min_value, max_time, max_value, s, crit_fs, crit_fs_der)

def cut_front(times, values, min_value, max_value, crit_fs, s):
    max_index = numpy.argmax(values >= max_value)

    if max_index == 0:
        #no point high enough was found so use the highest point
        s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(times, values, None, None)
        max_index = numpy.argmax(values)
        max_value = values[max_index]

        if min_value >= max_value:
            min_value = 1e-3 * max_value
    
    max_time = times[max_index]

    smooth_value = smoothing.smooth_data(times, values, crit_fs, s)

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, smooth_value, ext=1)

    error = abs(spline(max_time) - values[max_index])/max(values)

    if error > 1e-2:
        #result is truly garbage, this means the shape is so distorted compared to the real system that even the spline is not accurate
        #this almost always happens when the peak is so sharp it only spans a few time points
        #it also means the rest of the optimization is not needed
        return spline, numpy.zeros(len(times))
    
    def goal(time):
        if (time < times[0]) or (time > times[-1]):
            #values outside of the allowed time region are not allowed
            return 1e10
        return (spline(time)-max_value)**2
    
    result = scipy.optimize.minimize(goal, max_time, method='powell', tol=1e-5)

    old_max_time = max_time
    max_time = float(result.x)

    min_index = numpy.argmin(values[:max_index] <= min_value)
    min_time = times[min_index]

    if min_time >= max_time:
        #this happens when the peak is so sharp the front only spans 2 indexes
        while min_time >= max_time:
            if min_index > 0:
                min_index -= 1
                min_time = times[min_index]
            else:
                break
    
    def goal(time):
        if time > max_time:
            #don't search to the right of the max_time for a lower value
            return 1e10
        if time < times[0]:
            #don't search before the start time
            return 1e10
        return (spline(time) - min_value)**2
    
    result = scipy.optimize.minimize(goal, min_time, method='powell', tol=1e-5)

    min_time = float(result.x)

    needed_points = int( (max_time - min_time) * 10)
    new_times_spline = numpy.linspace(min_time, max_time, needed_points)
    new_values_spline = spline(new_times_spline)
    
    spline = scipy.interpolate.InterpolatedUnivariateSpline(new_times_spline, new_values_spline, ext=1)
    
    return spline, score.cut_zero(times, smooth_value, min_value, max_value)[0]


import sys

import matplotlib
matplotlib.use('Agg')

size = 20

matplotlib.rc('font', size=size)          # controls default text sizes
matplotlib.rc('axes', titlesize=size)     # fontsize of the axes title
matplotlib.rc('axes', labelsize=size)    # fontsize of the x and y labels
matplotlib.rc('xtick', labelsize=size)    # fontsize of the tick labels
matplotlib.rc('ytick', labelsize=size)    # fontsize of the tick labels
matplotlib.rc('legend', fontsize=size)    # legend fontsize
matplotlib.rc('figure', titlesize=size)  # fontsize of the figure title
matplotlib.rc('figure', autolayout=True)

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D  

from CADETMatch.cache import cache

from pathlib import Path
import pandas
import numpy
import scipy.interpolate
import itertools

from cadet import Cadet, H5
from addict import Dict

#parallelization
import multiprocessing

import os
import warnings
import CADETMatch.util as util
import logging
import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.smoothing as smoothing

saltIsotherms = {b'STERIC_MASS_ACTION', b'SELF_ASSOCIATION', b'MULTISTATE_STERIC_MASS_ACTION', 
                 b'SIMPLE_MULTISTATE_STERIC_MASS_ACTION', b'BI_STERIC_MASS_ACTION'}


from matplotlib.colors import ListedColormap
import matplotlib.cm
cmap = matplotlib.cm.winter

# Get the colormap colors
my_cmap = cmap(numpy.arange(cmap.N))

# Set alpha
my_cmap[:,-1] = 1.0

# Create new colormap
my_cmap = ListedColormap(my_cmap)

cm_plot = matplotlib.cm.gist_rainbow

def get_color(idx, max_colors, cmap):
    return cmap(1.*float(idx)/max_colors)

def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings['resultsDirLog'], "graph.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    fullGeneration = int(sys.argv[2])

    #full generation 1 = 2D, 2 = 2D + 3D

    graphMeta(cache, map_function)
    graphProgress(cache, map_function)

    if fullGeneration:
        graphSpace(fullGeneration, cache, map_function)
        graphExperiments(cache, map_function)    

def graphDistance(cache, map_function):

    resultDir = Path(cache.settings['resultsDir'])
    result_h5 = resultDir / "result.h5"

    output_distance = cache.settings['resultsDirSpace'] / "distance"
    output_distance_meta = output_distance / "meta"

    output_distance.mkdir(parents=True, exist_ok=True)

    output_distance_meta.mkdir(parents=True, exist_ok=True)

    parameter_headers_actual = cache.parameter_headers_actual
    score_headers = cache.score_headers
    meta_headers = cache.meta_headers

    if result_h5.exists():
        data = H5()
        data.filename = result_h5.as_posix()
        data.load()

        temp = []

        if 'mean' in data.root and 'confidence' in data.root and 'distance_correct' in data.root:
            for idx_parameter, parameter in enumerate(parameter_headers_actual):
                for idx_score, score in enumerate(score_headers):
                    temp.append( (output_distance, parameter, idx_parameter, score, idx_score, data.root.distance_correct[:,idx_parameter], data.root.output[:,idx_score],
                                    data.root.mean[:,idx_parameter], data.root.confidence[:,idx_parameter]) )

            for idx_parameter, parameter in enumerate(parameter_headers_actual):
                for idx_score, score in enumerate(meta_headers):
                    temp.append( (output_distance_meta, parameter, idx_parameter, score, idx_score, data.root.distance_correct[:,idx_parameter], data['output_meta'][:,idx_score],
                                data.root.mean[:,idx_parameter], data.root.confidence[:,idx_parameter]) )

            list(map_function(plot_2d_scatter, temp))

def plot_2d_scatter(args):
    output_directory, parameter, parameter_idx, score, score_idx, parameter_data, score_data, mean_data, confidence_data = args
    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(parameter_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(mean_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_mean.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)
    graph.scatter(confidence_data, score_data, c=score_data, cmap=my_cmap)
    graph.set_xlabel(parameter)
    graph.set_ylabel(score)
    filename = "%s_%s_confidence.png" % (parameter_idx, score_idx)
    fig.savefig(str(output_directory / filename))

def graphExperiments(cache, map_function):
    multiprocessing.get_logger().info("starting simulation graphs")
    directory = Path(cache.settings['resultsDirEvo'])

    #find all items in directory
    pathsH5 = directory.glob('*.h5')
    pathsPNG = directory.glob('*.png')

    #make set of items based on removing everything after _
    existsH5 = {str(path.name).split('_', 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split('_', 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(toGenerate, itertools.repeat(sys.argv[1]), itertools.repeat(cache.settings['resultsDirEvo']), itertools.repeat('%s_%s_EVO.png'))

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending simulation graphs")

def graphMeta(cache, map_function):
    multiprocessing.get_logger().info("starting meta graphs")
    directory = Path(cache.settings['resultsDirMeta'])

    #find all items in directory
    pathsH5 = directory.glob('*.h5')
    pathsPNG = directory.glob('*.png')

    #make set of items based on removing everything after _
    existsH5 = {str(path.name).split('_', 1)[0] for path in pathsH5}
    existsPNG = {str(path.name).split('_', 1)[0] for path in pathsPNG}

    toGenerate = existsH5 - existsPNG

    args = zip(toGenerate, itertools.repeat(sys.argv[1]), itertools.repeat(cache.settings['resultsDirMeta']), itertools.repeat('%s_%s_meta.png'))

    list(map_function(plotExperiments, args))
    multiprocessing.get_logger().info("ending meta graphs")

def graph_simulation_unit(simulation, unit, graph):
    ncomp = int(simulation.root.input.model[unit].ncomp)
    isotherm = bytes(simulation.root.input.model[unit].adsorption_model)

    #This does not work correctly since we could be reading an outlet or tubing that does not have an isotherm
    #hasSalt = isotherm in saltIsotherms

    solution_times = simulation.root.output.solution.solution_times

    comps = []

    hasColumn = any('column' in i for i in simulation.root.output.solution[unit].keys())
    hasPort = any('port' in i for i in simulation.root.output.solution[unit].keys())

    if hasColumn:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]['solution_column_outlet_comp_%03d' % i])
    elif hasPort:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]['solution_outlet_port_000_comp_%03d' % i])
    else:
        for i in range(ncomp):
            comps.append(simulation.root.output.solution[unit]['solution_outlet_comp_%03d' % i])

    #for now lets assume that if your first component is much higher concentration than the other components then the first component is salt
    max_comps = [max(comp) for comp in comps]

    hasSalt = any([max_comps[0] > (comp*10) for comp in max_comps[1:]])

    if hasSalt:
        graph.set_title("Output %s" % unit)
        graph.plot(solution_times, comps[0], 'b-', label="Salt")
        graph.set_xlabel('time (s)')
        
        # Make the y-axis label, ticks and tick labels match the line color.
        graph.set_ylabel('mM Salt', color='b')
        graph.tick_params('y', colors='b')

        axis2 = graph.twinx()
        for idx, comp in enumerate(comps[1:]):
            axis2.plot(solution_times, comp, '-', color=get_color(idx, len(comps) - 1, cm_plot), label="P%s" % idx)
        axis2.set_ylabel('mM Protein', color='r')
        axis2.tick_params('y', colors='r')


        lines, labels = graph.get_legend_handles_labels()
        lines2, labels2 = axis2.get_legend_handles_labels()
        axis2.legend(lines + lines2, labels + labels2, loc=0)
    else:
        graph.set_title("Output %s" % unit)
        
        #colors = ['r', 'g', 'c', 'm', 'y', 'k']
        for idx, comp in enumerate(comps):
            graph.plot(solution_times, comp, '-', color=get_color(idx, len(comps), cm_plot), label="P%s" % idx)
        graph.set_ylabel('mM Protein', color='r')
        graph.tick_params('y', colors='r')
        graph.set_xlabel('time (s)')

        lines, labels = graph.get_legend_handles_labels()
        graph.legend(lines, labels, loc=0)

def plotExperiments(args):
    save_name_base, json_path, directory, file_pattern = args
    if json_path != cache.json_path:
        cache.setup(json_path)

    target = cache.target
    settings = cache.settings
    for experiment in settings['experiments']:
        experimentName = experiment['name']
        
        dst = Path(directory, file_pattern % (save_name_base, experimentName))

        if dst.exists():
            #this item has already been plotted, this means that plots are occurring faster than generations are so kill this version
            sys.exit()

        units_used = cache.target[experimentName]['units_used']

        numPlotsSeq = [len(units_used)]
        #Shape and ShapeDecay have a chromatogram + derivative
        for feature in experiment['features']:
            if feature['type'] in ('Shape', 'ShapeDecay', 'ShapeFront', 'ShapeBack'):
                numPlotsSeq.append(2)
            elif feature['type'] in ('AbsoluteTime', 'AbsoluteHeight'):
                pass
            else:
                numPlotsSeq.append(1)

        numPlots = sum(numPlotsSeq)

        exp_time = target[experimentName]['time']
        exp_value = target[experimentName]['value']

        fig = figure.Figure(figsize=[24, 12*numPlots])
        canvas = FigureCanvas(fig)

        simulation = Cadet()
        h5_path = Path(directory) / (file_pattern.replace('png', 'h5') % (save_name_base, experimentName))
        simulation.filename = h5_path.as_posix()
        simulation.load()

        results = {}
        results['simulation'] = simulation

        for idx, unit in enumerate(units_used):
            graph_simulation_unit(simulation, unit, fig.add_subplot(numPlots, 1, idx+1))

        graphIdx = idx + 2
        for idx, feature in enumerate(experiment['features']):
            featureName = feature['name']
            featureType = feature['type']

            feat = target[experimentName][featureName]

            selected = feat['selected']
            exp_time = feat['time'][selected]
            exp_value = feat['value'][selected]

            sim_time, sim_value = get_times_values(simulation, target[experimentName][featureName])
            

            if featureType in ('similarity', 'similarityDecay', 'similarityHybrid', 'similarityHybrid2', 'similarityHybrid2_spline', 'similarityHybridDecay', 
                               'similarityHybridDecay2', 'curve', 'breakthrough', 'dextran', 'dextranHybrid', 'dextranHybrid2', 'dextranHybrid2_spline',
                               'similarityCross', 'similarityCrossDecay', 'breakthroughCross', 'SSE', 'LogSSE', 'breakthroughHybrid', 'breakthroughHybrid2',
                               'Shape', 'ShapeDecay', 'Dextran', 'DextranAngle', 'DextranTest', 'DextranQuad',
                               'Dextran3', 'DextranShape', 'ShapeDecaySimple', 'ShapeSimple', 'DextranSSE',
                               'ShapeFront', 'ShapeBack', 'ShapeNoDer', 'ShapeDecayNoDer'):
                
                graph = fig.add_subplot(numPlots, 1, graphIdx) #additional +1 added due to the overview plot
                graph.plot(sim_time, sim_value, 'r--', label='Simulation')
                graph.plot(exp_time, exp_value, 'g:', label='Experiment')
                graphIdx += 1
            
            if featureType in ('derivative_similarity', 'derivative_similarity_hybrid', 'derivative_similarity_hybrid2', 'derivative_similarity_cross', 'derivative_similarity_cross_alt',
                                 'derivative_similarity_hybrid2_spline', 'similarityHybridDecay2_spline',
                                 'Shape', 'ShapeDecay', 'ShapeFront', 'ShapeBack'):
                exp_spline = smoothing.smooth_data_derivative(exp_time, exp_value, feat['critical_frequency'], 
                                                              feat['smoothing_factor'], feat['critical_frequency_der'])
                sim_spline = smoothing.smooth_data_derivative(sim_time, sim_value, feat['critical_frequency'], 
                                                              feat['smoothing_factor'], feat['critical_frequency_der'])
                
                graph = fig.add_subplot(numPlots, 1, graphIdx) #additional +1 added due to the overview plot
                graph.plot(sim_time, sim_spline, 'r--', label='Simulation')
                graph.plot(exp_time, exp_spline, 'g:', label='Experiment')
                graphIdx += 1
            
            if featureType in ('fractionation', 'fractionationCombine', 'fractionationMeanVariance', 
                               'fractionationMoment', 'fractionationSlide', 'fractionationSSE'):
                cache.scores[featureType].run(results, feat)


                graph_exp = results['graph_exp']
                graph_sim = results['graph_sim']
                graph_sim_offset = results.get('graph_sim_offset', {})

                findMax = 0
                max_comp = {}
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max_value

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time, values) = zip(*value)
                    max_value = max(values)
                    findMax = max(findMax, max_value)
                    max_comp[key] = max(max_comp[key], max_value)

                graph = fig.add_subplot(numPlots, 1, graphIdx) #additional +1 added due to the overview plot
                factors = []
                for idx, (key, value) in enumerate(graph_sim.items()):
                    (time, values) = zip(*value)
                    values = numpy.array(values)
                    mult = 1.0
                    if max(values) < (.2 * findMax):
                        mult = findMax/(2 * max_comp[key])
                        values = values * mult
                    factors.append(mult)

                    time_offset = graph_sim_offset.get(key, None)
                    if time_offset is not None:
                        if time_offset >0:
                            time_str = "time early (s): %.3g" % abs(time_offset)
                        else:
                            time_str = "time late (s): %.3g" % abs(time_offset)
                    else:
                        time_str = ""

                    label = 'Simulation Comp: %s Mult:%.2f %s' % (key, mult, time_str)

                    graph.plot(time, values, '--', color=get_color(idx, len(graph_sim), cm_plot), label=label)

                for idx, (key, value) in enumerate(graph_exp.items()):
                    (time, values) = zip(*value)
                    values = numpy.array(values)
                    mult = factors[idx]
                    values = values * mult
                    graph.plot(time, values, ':', color=get_color(idx, len(graph_sim), cm_plot), label='Experiment Comp: %s Mult:%.2f' % (key, mult))
                graphIdx += 1
            graph.legend()
        fig.set_size_inches((24,12*numPlots))
        fig.savefig(str(dst))

def graphSpace(fullGeneration, cache, map_function):
    multiprocessing.get_logger().info("starting space graphs")
    progress_path = Path(cache.settings['resultsDirBase']) / "result.h5"

    output_2d = cache.settings['resultsDirSpace'] / "2d"
    output_3d = cache.settings['resultsDirSpace'] / "3d"

    output_2d.mkdir(parents=True, exist_ok=True)
    output_3d.mkdir(parents=True, exist_ok=True)
    
    results = H5()
    results.filename = progress_path.as_posix()

    results.load(paths=['/output', '/output_meta', '/is_extended_input'])

    if results.root.is_extended_input:
        results.load(paths=['/input_transform_extended'], update=True)
        input = results.root.input_transform_extended
    else:
        results.load(paths=['/input_transform'], update=True)
        input = results.root.input_transform

    output = results.root.output
    output_meta = results.root.output_meta

    input_headers = cache.parameter_headers
    output_headers = cache.score_headers
    meta_headers = cache.meta_headers

    input_indexes = list(range(len(input_headers)))
    output_indexes = list(range(len(output_headers)))
    meta_indexes = list(range(len(meta_headers)))

    comp_two = list(itertools.combinations(input_indexes, 2))
    comp_one = list(itertools.combinations(input_indexes, 1))

    #2d plots
    if fullGeneration >= 1:
        multiprocessing.get_logger().info("starting 2d space graphs")
        graphDistance(cache, map_function)

        seq = []
        for (x,), y in itertools.product(comp_one, output_indexes):
            seq.append( [output_2d.as_posix(), input_headers[x], output_headers[y], input[:,x], output[:,y]] )

        for (x,), y in itertools.product(comp_one, meta_indexes):
            seq.append( [output_2d.as_posix(), input_headers[x], meta_headers[y], input[:,x], output_meta[:,y]] )
        list(map_function(plot_2d, seq))
        multiprocessing.get_logger().info("ending 2d space graphs")

    #3d plots
    if fullGeneration >= 2:
        multiprocessing.get_logger().info("starting 3d space graphs")
        seq = []
        for (x, y), z in itertools.product(comp_two, output_indexes):
            seq.append( [output_3d.as_posix(), input_headers[x], input_headers[y], output_headers[z], input[:,x], input[:,y], output[:,z]] )

        for (x, y), z in itertools.product(comp_two, meta_indexes):
            seq.append( [output_3d.as_posix(), input_headers[x], input_headers[y], meta_headers[z], input[:,x], input[:,y], output_meta[:,z]] )
        list(map_function(plot_3d, seq))
        multiprocessing.get_logger().info("ending 3d space graphs")
    multiprocessing.get_logger().info("ending space graphs")

def plot_3d(arg):
    "This leaks memory and is run in a separate short-lived process, do not integrate into the matching main process"
    directory_path, header1, header2, header3, data1, data2, scores = arg
    directory = Path(directory_path)

    scoreName = header3
    if scoreName == 'SSE':
        scores = -numpy.log10(scores)
        scoreName = '-log10(%s)' % scoreName
    
    x = data1
    y = data2

    fig = figure.Figure()
    canvas = FigureCanvas(fig)
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(numpy.log10(x), numpy.log10(y), scores, c=scores, cmap=my_cmap)
    ax.set_xlabel('log10(%s)' % header1)
    ax.set_ylabel('log10(%s)' % header2)
    ax.set_zlabel(scoreName)
    filename = "%s_%s_%s.png" % (header1, header2, scoreName)
    filename = filename.replace(':', '_').replace('/', '_')
    fig.savefig(str(directory / filename))
    
def plot_2d(arg):
    directory_path, header_x, scoreName, data, scores = arg

    multiprocessing.get_logger().info("2d graph %s %s start" % (header_x, scoreName))

    plot_2d_single(directory_path, header_x, scoreName, data, scores)

    plot_2d_single(directory_path, header_x, '1- ' + scoreName, data, 1-scores)

    multiprocessing.get_logger().info("2d graph %s %s stop" % (header_x, scoreName))
    

def plot_2d_single(directory_path, header_x, scoreName, data, scores):
    directory = Path(directory_path)

    if numpy.all(scores <= 0):
        scores = numpy.abs(scores)

    #score_scale = numpy.max(scores)/numpy.min(scores)

    if  scoreName.startswith('1-'):
        keep = scores > 0
        scores = numpy.log10(scores[keep])
        data = data[keep]
        scoreName = 'log10(%s)' % scoreName

    fig = figure.Figure(figsize=[10,10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    format = '%s'
    if numpy.max(data)/numpy.min(data[data>0]) > 100.0:
        keep = data>0
        data = numpy.log10(data[keep])
        scores = scores[keep]
        format = 'log10(%s)'

    data_norm = (data - numpy.min(data))/(numpy.max(data) - numpy.min(data))
    scores_norm = (scores - numpy.min(scores))/(numpy.max(scores) - numpy.min(scores))
    scatter_data = numpy.array([data_norm,scores_norm]).T
    scatter_data_round = numpy.round(scatter_data, 2)
    scatter_data_round, idx, counts = numpy.unique(scatter_data_round, axis=0, return_index=True, return_counts=True)

    counts_norm =  (counts - numpy.min(counts))/(numpy.max(counts) - numpy.min(counts))*100+10

    graph.scatter(data[idx], scores[idx], c=scores[idx], s=counts_norm, cmap=cmap)

    graph.set_xlabel(format % header_x)
    graph.set_ylabel(scoreName)
    graph.set_xlim(min(data), max(data), auto=True)
    filename = "%s_%s.png" % (header_x, scoreName)
    filename = filename.replace(':', '_').replace('/', '_')

    fig.savefig(str(directory / filename))

def graphProgress(cache, map_function):
    multiprocessing.get_logger().info("starting progress graphs")
    results = Path(cache.settings['resultsDirBase'])
    progress = results / "progress.csv"
    
    df = pandas.read_csv(progress)

    output = cache.settings['resultsDirProgress']

    x = ['Generation',]
    y = ['Meta Front', 'Meta Min', 'Meta Product', 'Meta Mean', 'Meta SSE']

    temp = []
    for x,y in itertools.product(x,y):
        if x != y:
            temp.append( (df[[x,y]], output) )

    list(map_function(singleGraphProgress, temp))
    multiprocessing.get_logger().info("ending progress graphs")

def singleGraphProgress(arg):
    df, output = arg

    i,j = list(df)

    fig = figure.Figure(figsize=(10,10))
    canvas = FigureCanvas(fig)

    graph = fig.add_subplot(1, 1, 1)

    graph.plot(df[i],df[j])
    a = max(df[j])
    graph.set_ylim((0,1.1*max(df[j])))
    graph.set_title('%s vs %s' % (i,j))
    graph.set_xlabel(i)
    graph.set_ylabel(j)

    filename = "%s vs %s.png" % (i,j)
    file_path = output / filename
    fig.savefig(str(file_path))



    fig = figure.Figure(figsize=(10,10))
    canvas = FigureCanvas(fig)

    graph = fig.add_subplot(1, 1, 1)

    graph.plot(df[i],numpy.log10(1-df[j]))
    a = max(df[j])
    #graph.set_ylim((0,1.1*max(df[j])))
    graph.set_title('%s vs log10(1-%s)' % (i,j))
    graph.set_xlabel(i)
    graph.set_ylabel('log10(1-%s)' % j)

    filename = "%s vs log10(1-%s).png" % (i,j)
    file_path = output / filename
    fig.savefig(str(file_path))


def get_times_values(simulation, target, selected = None):

    times = simulation.root.output.solution.solution_times

    isotherm = target['isotherm']

    if isinstance(isotherm, list):
        values = numpy.sum([simulation[i] for i in isotherm], 0)
    else:
        values = simulation[isotherm]
    
    if selected is None:
        selected = target['selected']

    return times[selected], values[selected] * target['factor']

if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import setuptools

from CADETMatch import version

with open("README.md", "r") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name=version.__title__,
    version=version.__version__,
    author=version.__author__,
    author_email=version.__email__,
    description=version.__summary__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/modsim/CADET-Match",
    packages=setuptools.find_packages(),
    install_requires=[
          'joblib>=0.14.0',
          'addict>=2.2.1',
          'emcee>=3',
          'SAlib',
          'corner>=2.0.1',
          'deap>=1.3.0',
          'psutil>=5.6.3',
          'openpyxl>=3.0.0,!=3.0.2',
          'numpy>=1.16.2',
          'scipy>=1.3.1',
          'matplotlib>=3.1.1',
          'pandas>=0.25.1',
          'h5py>=2.10.0',
          'CADET>=0.3',
          'seaborn>=0.9',
          'sklearn',
          'scikit-learn>=0.22',
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
) 


import CADETMatch.util as util
import CADETMatch.calc_coeff as calc_coeff
import CADETMatch.score as score
from addict import Dict

name = "Ceiling"

def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 0
    settings.meta_mask = True
    settings.count = 1
    return settings

def run(sim_data, feature):
    "similarity, value, start stop"
    selected = feature['selected']
    exp_data_values = feature['value'][selected]
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)
    
    max_value = max(sim_data_values)

    if max_value <= feature['max_value']:
        value = 1.0
    else:
        value = feature['value_function'](max_value)

    temp = [value,]

    return (temp, util.sse(sim_data_values, exp_data_values), len(sim_data_values), 
            sim_time_values, sim_data_values, exp_data_values, [1.0 - i for i in temp])

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    temp['value_function'] = score.value_function(feature['max_value'], abstol)
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_Height" % name]
    return temp





import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import numpy
import numpy.linalg
from addict import Dict
import sys
import CADETMatch.smoothing as smoothing
import multiprocessing

name = "DextranSSE"

def get_settings(feature):
    settings = Dict()
    settings.adaptive = False
    settings.badScore = -sys.float_info.max
    settings.meta_mask = True
    settings.count = 1
    return settings

def run(sim_data, feature):
    "special score designed for dextran. This looks at only the front side of the peak up to the maximum slope and pins a value at the elbow in addition to the top"
    exp_time_values = feature['time']

    min_value = feature['min_value']
    max_value = feature['max_value']

    selected = feature['selected']
        
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)

    if max(sim_data_values) < max_value: #the system has no point higher than the value we are looking for
        #remove hard failure
        max_value = max(sim_data_values)

    exp_time_values = exp_time_values[selected]
    exp_data_zero = feature['exp_data_zero']

    sim_data_values_smooth = smoothing.smooth_data(sim_time_values, sim_data_values, 
                                                    feature['critical_frequency'], feature['smoothing_factor'])
    sim_data_zero, sim_min_time, sim_min_value, sim_max_time, sim_max_value = score.cut_zero(sim_time_values, sim_data_values_smooth, min_value, max_value)

    sse = util.sse(sim_data_zero, exp_data_zero)

    data = [-sse,], sse, len(sim_data_zero), sim_time_values, sim_data_zero, exp_data_zero, [sse,]

    return data

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    #change the stop point to be where the max positive slope is along the searched interval
    name = '%s_%s' % (sim.root.experiment_name,   feature['name'])
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(selectedTimes, selectedValues, name, cache)
    values_der = smoothing.smooth_data_derivative(selectedTimes, selectedValues, crit_fs, s, crit_fs_der)
    
    smooth_value = smoothing.smooth_data(selectedTimes, selectedValues, crit_fs, s)

    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(selectedTimes, values_der, ext=1)
    spline = scipy.interpolate.InterpolatedUnivariateSpline(selectedTimes, smooth_value, ext=1)

    min_time, min_value, max_time, max_value = score.find_cuts(selectedTimes, smooth_value, spline, spline_der)

    exp_data_zero, exp_min_time, exp_min_value, exp_max_time, exp_max_value = score.cut_zero(selectedTimes, smooth_value, min_value, max_value)

    if exp_min_time is not None:
        min_time = exp_min_time
        min_value = exp_min_value
        max_time = exp_max_time
        max_value = exp_max_value

    multiprocessing.get_logger().info("Dextran %s  start: %s   stop: %s  max value: %s", name, 
                                      min_time, max_time, max_value)

    temp['min_time'] = feature['start']
    temp['max_time'] = feature['stop']
    temp['min_value'] = min_value
    temp['max_value'] = max_value
    temp['exp_data_zero'] = exp_data_zero
    temp['offsetTimeFunction'] = score.time_function_decay_cv(CV_time, selectedTimes, max_time)
    temp['peak_max'] = max_value
    temp['smoothing_factor'] = s
    temp['critical_frequency'] = crit_fs
    temp['critical_frequency_der'] = crit_fs_der
    temp['smooth_value'] = smooth_value
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_SSE" % name,]
    return temp


#This module simulates noise (systemic and random) to use for kernel density estimation based on experimental data

#Currently the randomness is

#Pump flow rate
#Pump delay
#Base noise
#Signal noise

import itertools
import CADETMatch.score as score
from cadet import Cadet, H5
from pathlib import Path
import numpy
import multiprocessing
from sklearn.neighbors import KernelDensity
from sklearn.model_selection import cross_val_score
from sklearn import preprocessing
import copy

bw_tol = 1e-4

import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py

import scipy.optimize

import CADETMatch.util as util
import CADETMatch.cache as cache

import CADETMatch.synthetic_error as synthetic_error

import joblib
import subprocess
import sys

atol = 1e-3
rtol = 1e-3

def bandwidth_score(bw, data, store):
    bandwidth = 10**bw[0]
    kde_bw = KernelDensity(kernel='gaussian', bandwidth=bandwidth, atol=atol, rtol=rtol)
    scores = cross_val_score(kde_bw, data, cv=3)
    mean = -numpy.mean(scores)
    store.append( [bandwidth, mean] )
    return mean

def get_bandwidth(scores, cache):
    store = []
    
    bw_sample = numpy.linspace(-3, 0, 20)
    
    bw_score = [bandwidth_score([bw,], scores, store) for bw in bw_sample]
    
    idx = numpy.argmin(bw_score)
    
    bw_start = bw_sample[idx]

    result = scipy.optimize.minimize(bandwidth_score, bw_start, args = (scores,store,), method='powell')
    bandwidth = 10**result.x
    multiprocessing.get_logger().info("selected bandwidth %s", bandwidth)

    store = numpy.array(store)
    return bandwidth, store

def mirror(data):
    data_max = numpy.max(data,0)
    
    mirror_index = data_max <= 1.0
    keep_index = data_max > 1.0
    
    data_min = data_max - data    
    data_mask = numpy.ma.masked_equal(data_min, 0.0, copy=False)
    min_value = data_mask.min(axis=0)
    
    data_mirror = numpy.zeros(data.shape)
    
    data_mirror[:,mirror_index] = data_max[mirror_index] + data_max[mirror_index] - numpy.copy(data[:,mirror_index]) + min_value[mirror_index]
    data_mirror[:,keep_index] = data[:,keep_index]
    full_data = numpy.vstack([data_mirror, data])
    
    return full_data

def setupKDE(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler, cache)

    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth, atol=bw_tol).fit(scores_scaler)

    joblib.dump(scaler, mcmcDir / 'kde_scaler.joblib')

    joblib.dump(kde, mcmcDir / 'kde_score.joblib')

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.root.bandwidth = bandwidth
    h5_data.root.store = numpy.array(store)
    h5_data.root.scores = scores
    h5_data.root.scores_mirror = scores_mirror
    h5_data.root.scores_mirror_scaled = scores_scaler
    h5_data.save()

    cwd = str(Path(__file__).parent)
    ret = subprocess.run([sys.executable, 'graph_kde.py', str(cache.json_path), str(util.getCoreCounts())], 
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    util.log_subprocess('graph_kde.py', ret)

    return kde, scaler

def getScaler(data):
    scaler = preprocessing.StandardScaler().fit(data)
    return scaler

def getKDE(cache):
    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    kde = joblib.load(mcmcDir / 'kde_score.joblib')

    scaler = joblib.load(mcmcDir / 'kde_scaler.joblib')

    return kde, scaler

def getKDEPrevious(cache):
    if 'mcmc_h5' in cache.settings:
        mcmc_h5 = Path(cache.settings['mcmc_h5'])
        mcmcDir = mcmc_h5.parent

        if mcmcDir.exists():
            kde = joblib.load(mcmcDir / 'kde_prior.joblib')

            scaler = joblib.load(mcmcDir / 'kde_prior_scaler.joblib')

            return kde, scaler
    return None, None

def generate_data(cache):
    scores = generate_synthetic_error(cache)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    save_scores = mcmcDir / 'scores_used.npy'

    numpy.save(save_scores, scores)

    scores_mirror = mirror(scores)

    scaler = getScaler(scores_mirror)

    scores_scaler = scaler.transform(scores_mirror)

    bandwidth, store = get_bandwidth(scores_scaler, cache)

    return scores, bandwidth

def synthetic_error_simulation(json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings['resultsDirLog'], "main.log")
        cache.cache.setup(json_path)

    scores = []
    outputs = {}
    simulations = {}
    experiment_failed = False
    
    delays_store = []
    flows_store = []
    load_store = []
    uv_store = {}

    for experiment in cache.cache.settings['kde_synthetic']:
        delay_settings = experiment['delay']
        flow_settings = experiment['flow']
        load_settings = experiment['load']
        experimental_csv = experiment['experimental_csv']
        uv_noise = experiment['uv_noise']
        units = experiment['units']
        name = experiment['name']

        data = numpy.loadtxt(experimental_csv, delimiter=',')
        times = data[:,0]

        if 'file_path' in experiment:
            template_path = Path(experiment['file_path'])
        else:
            resultsDir = cache.cache.settings['resultsDir']
            if 'resultsDirOriginal' in cache.cache.settings:
                resultsDir = Path(cache.cache.settings['resultsDirOriginal'])

            template_path = resultsDir / "misc" / ("template_%s_base.h5" % name)

        temp = Cadet()
        temp.filename = template_path.as_posix()
        temp.load()

        util.setupSimulation(temp, times, cache.cache.target[name]['smallest_peak'], cache.cache)

        nsec = temp.root.input.solver.sections.nsec

        for unit in units:
            unit_name = 'unit_%03d' % unit
            for comp in range(temp.root.input.model[unit_name].ncomp):
                comp_name = 'solution_outlet_comp_%03d' % comp
                error_uv = numpy.random.normal(uv_noise[0], uv_noise[1], len(temp.root.input.solver.user_solution_times))
                uv_store['%s_%s_%s' % (name, unit, comp)] = error_uv

        def post_function(simulation):
            #baseline drift need to redo this
            #error_slope = numpy.random.normal(error_slope_settings[0], error_slope_settings[1], 1)[0]

            for unit in units:
                unit_name = 'unit_%03d' % unit
                for comp in range(simulation.root.input.model[unit_name].ncomp):
                    comp_name = 'solution_outlet_comp_%03d' % comp
                    error_uv = uv_store['%s_%s_%s' % (name, unit, comp)]
                    simulation.root.output.solution[unit_name][comp_name] = simulation.root.output.solution[unit_name][comp_name] + error_uv
    
        error_delay = Cadet(temp.root)

        delays = numpy.random.uniform(delay_settings[0], delay_settings[1], nsec)
        delays_store.extend(delays)
    
        synthetic_error.pump_delay(error_delay, delays)

        flow = numpy.random.normal(flow_settings[0], flow_settings[1], error_delay.root.input.solver.sections.nsec)
        flows_store.extend(flow)
    
        synthetic_error.error_flow(error_delay, flow)
    
        load = numpy.random.normal(load_settings[0], load_settings[1], error_delay.root.input.solver.sections.nsec)
        load_store.extend(load)
    
        synthetic_error.error_load(error_delay, load)

        exp_info = None
        for exp in cache.cache.settings['experiments']:
            if exp['name'] == name:
                exp_info = exp
                break

        result = util.runExperiment(None, exp_info, cache.cache.settings, cache.cache.target, error_delay, error_delay.root.timeout, cache.cache, post_function=post_function)

        if result is not None:
            scores.extend(result['scores'])

            simulations[name] = result['simulation']

            for unit in units:
                unit_name = 'unit_%03d' % unit
                for comp in range(result['simulation'].root.input.model[unit_name].ncomp):
                    outputs['%s_unit_%03d_comp_%03d' % (name, int(unit), comp)] = result['simulation'].root.output.solution['unit_%03d' % int(unit)]['solution_outlet_comp_%03d' % comp]
        else:
            experiment_failed = True

    if experiment_failed:
        return None, None, None, None

    errors = {}
    errors['delays'] = delays_store
    errors['flows'] = flows_store
    errors['load'] = load_store
    errors['uv_store'] = uv_store

    return scores, simulations, outputs, errors

def generate_synthetic_error(cache):
    if 'kde_synthetic' in cache.settings:
        count_settings = int(cache.settings['kde_synthetic'][0]['count'])
        
        scores_all = []
        times = {}
        outputs_all = {}
        
        delays_all = []
        flows_all = []
        load_all = []
        uv_store_all = {}

        for scores, simulations, outputs, errors in cache.toolbox.map(synthetic_error_simulation, [cache.json_path] * count_settings):
            if scores and simulations and outputs:

                delays_all.append(errors['delays'])
                flows_all.append(errors['flows'])
                load_all.append(errors['load'])

                for key,value in errors['uv_store'].items():
                    uv = uv_store_all.get(key, [])
                    uv.append(value)
                    uv_store_all[key] = uv 

                scores_all.append(scores)

                for key,value in outputs.items():
                    temp = outputs_all.get(key, [])
                    temp.append(value)
                    outputs_all[key] = temp

                for key,sim in simulations.items():
                    if key not in times:
                        times[key] = sim.root.output.solution.solution_times
                                        
        scores = numpy.array(scores_all)

        dir_base = cache.settings.get('resultsDirBase')
        file = dir_base / 'kde_data.h5'

        kde_data = H5()
        kde_data.filename = file.as_posix()

        kde_data.root.scores = scores

        for output_name, output in outputs_all.items():
            kde_data.root[output_name] = numpy.array(output)

        for time_name, time in times.items():
            kde_data.root['%s_time' % time_name] = time

        kde_data.root.errors.flows = convert_to_array(flows_all)
        kde_data.root.errors.delays = convert_to_array(delays_all)
        kde_data.root.errors.load = convert_to_array(load_all)

        for key,value in uv_store_all.items():
            kde_data.root.errors[key] = numpy.array(value)

        kde_data.save()
                               
        return scores

    return None, None

def convert_to_array(seq):
    "this converts a sequence of arrays to a single numpy array and pads uneven rows with 0"
    max_len = numpy.max([len(a) for a in seq])
    return numpy.asarray([numpy.pad(a, (0, max_len - len(a)), 'constant', constant_values=0) for a in seq])


import numpy
import numpy.linalg
import scipy
import scipy.signal
import CADETMatch.util as util
import warnings

import multiprocessing

from cadet import H5

butter_order = 3

def refine_butter(times, values, x, y, fs, start):
    x = numpy.array(x)
    y = numpy.array(y)

    y_min = y - min(y)

    p3 = numpy.array([x, y]).T
    p1 = p3[0,:]
    p2 = p3[-1,:]
    
    def goal(crit_fs):
        crit_fs = 10.0**crit_fs[0]
        try:
            sos = scipy.signal.butter(butter_order, crit_fs, btype='lowpass', analog=False, fs=fs, output="sos")
        except ValueError:
            return 1e6
        low_passed = scipy.signal.sosfiltfilt(sos, values)
        
        sse = numpy.sum( (low_passed - values)**2 )
        
        pT = numpy.array([crit_fs, numpy.log(sse)]).T

        d = numpy.cross(p2-p1,p1-pT)/numpy.linalg.norm(p2-p1)
        
        return d
    
    start = numpy.log10(start)
    lb = numpy.log10(x[-1])
    ub = numpy.log10(x[0])
    diff = min(start - lb, ub-start)
    
    lb = start - 0.2 * diff
    ub = start + 0.2 * diff

    result_evo = scipy.optimize.differential_evolution(goal, ((lb, ub),), polish=False)
    
    crit_fs = 10**result_evo.x[0]
    
    return crit_fs

def refine_smooth(times, values, x, y, start, name):
    if name is None:
        name = 'unknown'

    x = numpy.array(x)
    y = numpy.array(y)

    y_min = y - min(y)

    p3 = numpy.array([x, y_min]).T
    p1 = p3[0,:]
    p2 = p3[-1,:]
    
    def goal(s):
        s = 10**s[0]
        with warnings.catch_warnings():
            warnings.filterwarnings('error')

            try:
                spline = scipy.interpolate.UnivariateSpline(times, values, s=s, k=5, ext=3)
            except Warning:
                multiprocessing.get_logger().info("caught a warning for %s %s", name, s)
                return 1e6
        
        pT = numpy.array([s, len(spline.get_knots())]).T

        d = numpy.cross(p2-p1,p1-pT)/numpy.linalg.norm(p2-p1)
        
        return d
    
    start = numpy.log10(start)
    lb = numpy.log10(x[-1])
    ub = numpy.log10(x[0])
    diff = min(start - lb, ub-start)
    
    lb = start - 0.2 * diff
    ub = start + 0.2 * diff

    result_evo = scipy.optimize.differential_evolution(goal, ((lb, ub),), polish=False)
    
    s = 10**result_evo.x[0]
    
    spline = scipy.interpolate.UnivariateSpline(times, values, s=s, k=5, ext=3)
    
    return s, len(spline.get_knots())

def find_butter(times, values):
    filters = []
    sse = []
    
    fs = 1.0/(times[1] - times[0])

    for i in numpy.logspace(1, -4, 50):
        try:
            sos = scipy.signal.butter(butter_order, i, btype='lowpass', analog=False, fs=fs, output="sos")
        except ValueError:
            continue
        low_passed = scipy.signal.sosfiltfilt(sos, values)

        filters.append(i)
        sse.append(  numpy.sum( (low_passed - values)**2 ) )
        
    L_x, L_y = util.find_Left_L(filters, numpy.log(sse))

    if L_x is not None:
        L_x = refine_butter(times, values, filters, numpy.log(sse), fs, L_x)
  
    return L_x

def smoothing_filter_butter(times, values, crit_fs):
    if crit_fs is None:
        return values
    fs = 1.0/(times[1] - times[0])

    sos = scipy.signal.butter(butter_order, crit_fs, btype='lowpass', analog=False, fs=fs, output="sos")
    low_passed = scipy.signal.sosfiltfilt(sos, values)
    return low_passed

def load_data(name, cache):
    crit_fs = None
    crit_fs_der = None
    s = None
    s_knots = 0

    #quick abort
    if name is None or cache is None:
        return s, crit_fs, crit_fs_der

    factor_file = cache.settings['resultsDirMisc'] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load()

    if name in data.root:
        s = float(data.root[name].s)

        crit_fs = data.root[name].crit_fs
        if crit_fs == -1.0:
            crit_fs = None

        crit_fs_der = data.root[name].crit_fs_der
        if crit_fs_der == -1.0:
            crit_fs_der = None

        s_knots = int(data.root[name].s_knots)
    else:
        return s, crit_fs, crit_fs_der

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info("smoothing_factor %s  %.3e  %s  %s knots %d", name, s, crit_fs_message, crit_fs_der_message, s_knots)

    return s, crit_fs, crit_fs_der


def find_smoothing_factors(times, values, name, cache):
    min = 1e-2

    s, crit_fs, crit_fs_der = load_data(name, cache)

    if s is not None:
        return s, crit_fs, crit_fs_der

    #normalize the data
    values = values * 1.0/max(values)
    
    crit_fs = find_butter(times, values)

    if crit_fs is None:
        multiprocessing.get_logger().info("%s butter filter disabled, no viable L point found", name)
    
    values_filter = smoothing_filter_butter(times, values, crit_fs)    

    spline = scipy.interpolate.UnivariateSpline(times, values_filter, s=min, k=5, ext=3)
    knots = []
    all_s = []

    knots.append(len(spline.get_knots()))
    all_s.append(min)

    #This limits to 1e-14 max smoothness which is way beyond anything normal
    for i in range(1,200):  
        s = min/(1.1**i)
        with warnings.catch_warnings():
            warnings.filterwarnings('error')

            try:
                spline = scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3)
                knots.append(len(spline.get_knots()))
                all_s.append(s)

                if len(spline.get_knots()) > 600:
                    break

            except Warning:
                multiprocessing.get_logger().info("caught a warning for %s %s", name, s)
                break
    
    knots = numpy.array(knots)
    all_s = numpy.array(all_s)
    
    s, s_knots = util.find_Left_L(all_s,knots)

    if s is not None:
        s, s_knots = refine_smooth(times, values_filter, all_s, knots, s, name)


    spline, factor = create_spline(times, values, crit_fs, s)
    
    #run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)
    values_filter = spline.derivative()(times) / factor
    crit_fs_der = find_butter(times, values_filter)

    record_smoothing(s, s_knots, crit_fs, crit_fs_der, knots, all_s, name, cache)
    
    return s, crit_fs, crit_fs_der

def record_smoothing(s, s_knots, crit_fs, crit_fs_der, knots, all_s, name=None, cache=None):
    if name is None or cache is None:
        return
    factor_file = cache.settings['resultsDirMisc'] / "find_smoothing_factor.h5"

    data = H5()
    data.filename = factor_file.as_posix()

    if factor_file.exists():
        data.load()

    if name not in data.root:
        data.root[name].knots = knots
        data.root[name].all_s = all_s
        data.root[name].s = float(s)
        data.root[name].s_knots = int(s_knots)
        if crit_fs is None:
            data.root[name].crit_fs = -1.0
        else:
            data.root[name].crit_fs = float(crit_fs)
        if crit_fs_der is None:
            data.root[name].crit_fs_der = -1.0
        else:
            data.root[name].crit_fs_der = float(crit_fs)
        data.save()

    crit_fs_message = "critical frequency disable"
    if crit_fs is not None:
        crit_fs_message = "critical frequency %.3e" % crit_fs

    crit_fs_der_message = "critical frequency der disable"
    if crit_fs_der is not None:
        crit_fs_der_message = "critical frequency der %.3e" % crit_fs_der

    multiprocessing.get_logger().info("smoothing_factor %s  %.3e  %s  %s knots %d", name, s, crit_fs_message, crit_fs_der_message, s_knots)

def create_spline(times, values, crit_fs, s):
    factor = 1.0/max(values)
    values = values * factor
    values_filter = smoothing_filter_butter(times, values, crit_fs)

    return scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3), factor

def smooth_data(times, values, crit_fs, s):
    spline, factor = create_spline(times, values, crit_fs, s)
    
    return spline(times) / factor

def smooth_data_derivative(times, values, crit_fs, s, crit_fs_der, smooth=True):
    spline, factor = create_spline(times, values, crit_fs, s)
    
    #run a quick butter pass to remove high frequency noise in the derivative (needed for some experimental data)
    values_filter = spline.derivative()(times) / factor
    if smooth:
        values_filter = butter(times, values_filter, crit_fs_der)
    return values_filter

def butter(times, values, crit_fs_der):
    factor = 1.0/max(values)
    values = values * factor
        
    values_filter = smoothing_filter_butter(times, values, crit_fs_der) / factor
    
    return values_filter



import sys

import matplotlib
matplotlib.use('Agg')
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from CADETMatch.cache import cache

from pathlib import Path
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py
import numpy
import scipy.stats

def main():
    cache.setup(sys.argv[1])
    cache.progress_path = Path(cache.settings['resultsDirBase']) / "progress.csv"

    generation = sys.argv[2]

    result_path = Path(cache.settings['resultsDirBase']) / "result.h5"
    output_spear = cache.settings['resultsDirSpace'] / "spearman"

    output_spear.mkdir(parents=True, exist_ok=True)

    with h5py.File(result_path, 'r') as h5:
        data_output = h5['/output'][()]
        data_generation = h5['/generation'][()]
        
        data_generation = h5['/generation'][(-1, -1)]
        
        
        data_output = h5['/output'][-data_generation:, :]
        data_in = h5['/input'][-data_generation:, :]
        data_meta = h5['/output_meta'][-data_generation:, :]
        

        data = numpy.concatenate([data_in, data_output, data_meta], axis=1)

        spear, p_value = scipy.stats.spearmanr(data)

        fig = figure.Figure()
        canvas = FigureCanvas(fig)
        graph = fig.add_subplot(1, 1, 1)
        im = graph.imshow(spear, cmap='RdGy', interpolation='nearest', vmin=-1, vmax=1)
        graph.figure.colorbar(im, ax=graph)
        filename = "generation_%s.png" % (generation)
        fig.savefig(str(output_spear / filename), bbox_inches='tight')

if __name__ == "__main__":
    main()




import CADETMatch.util as util
import numpy
from CADETMatch.abstract.transform import AbstractTransform

class VolumeLengthTransform(AbstractTransform):
    @property
    def name(self):
        return "volume_length"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    def transform(self):
        def trans_volume(i):
            return i

        def trans_length(i):
            return i

        return [trans_volume, trans_length]

    grad_transform = transform

    def untransform(self, seq):
        values = [seq[0]/seq[1], seq[1]]
        headerValues = [values[0], values[1], values[0]*values[1]]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:,0] = values[:,0] / values[:,1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:,0] = matrix[:,0]
        values[:,1] = matrix[:,1]
        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)
    
        if self.parameter.get('experiments', None) is None or experiment['name'] in self.parameter['experiments']:
            area_location = self.parameter['area_location']
            length_location = self.parameter['length_location']

            sim[area_location.lower()] = values[0]
            sim[length_location.lower()] = values[1]

        return values, headerValues

    def getBounds(self):
        minVolume = self.parameter['minVolume']
        maxVolume = self.parameter['maxVolume']
        minLength = self.parameter['minLength']
        maxLength = self.parameter['maxLength']

        minValues = [minVolume, minLength]
        maxValues = [maxVolume, maxLength]

        return minValues, maxValues

    def getGradBounds(self):
        return self.getBounds()

    def getHeaders(self):
        headers = []
        headers.append("Area")
        headers.append("Length")
        headers.append("Volume")
        return headers

    def getHeadersActual(self):
        headers = []
        headers.append("Volume")
        headers.append("Length")
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter['minVolume'] = lb[2]
        parameter['maxVolume'] = ub[2]
        parameter['minLength'] = lb[1]
        parameter['maxLength'] = ub[1]

class NormVolumeLengthTransform(VolumeLengthTransform):
    @property
    def name(self):
        return "norm_volume_length"

    def transform(self):
        minVolume = self.parameter['minVolume']
        maxVolume = self.parameter['maxVolume']
        minLength = self.parameter['minLength']
        maxLength = self.parameter['maxLength']

        def trans_volume(i):
            return (i - minVolume)/(maxVolume - minVolume)

        def trans_length(i):
            return (i - minLength)/(maxLength - minLength)

        return [trans_volume, trans_length]

    grad_transform = transform

    def untransform(self, seq):
        minVolume = self.parameter['minVolume']
        maxVolume = self.parameter['maxVolume']
        minLength = self.parameter['minLength']
        maxLength = self.parameter['maxLength']

        minValues = numpy.array([minVolume, minLength])
        maxValues = numpy.array([maxVolume, maxLength])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues

        values = [values[0]/values[1], values[1]]
        headerValues = [values[0], values[1], values[0]*values[1]]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        values = self.untransform_matrix_inputorder(matrix)
        values[:,0] = values[:,0] / values[:,1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minVolume = self.parameter['minVolume']
        maxVolume = self.parameter['maxVolume']
        minLength = self.parameter['minLength']
        maxLength = self.parameter['maxLength']

        minValues = numpy.array([minVolume, minLength])
        maxValues = numpy.array([maxVolume, maxLength])

        values = (maxValues - minValues) * matrix + minValues

        return values

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter['min'],], [self.parameter['max'],]

plugins = {"norm_volume_length": NormVolumeLengthTransform, "volume_length": VolumeLengthTransform}